#include "minunit.h"
#include <lcthw/bstrlib.h>
#include <stdarg.h>

static char *cstring = "foo-bar-baz";
static int cstring_len = 10;
static bstring b;

/**
 * Calling the same function to deallocate every
 * bstring is annoying, so here's a function
 * to deallocate them all with function arguments
 */
static int bdestroy_all(int count, ...)
{
	va_list bstrs;
	int i;
	int rc;

	va_start(bstrs, count);

	for(i = 0; i < count; i++)
	{
		bstring next = va_arg(bstrs, bstring);
		rc = bdestroy(next);
		if (rc != BSTR_OK)
		{
			return rc;
		}
	}

	va_end(bstrs);

	return BSTR_OK;
}

char *test_bfromcstr()
{
	mu_assert(cstring != NULL, "Invalid c string");
	b = bfromcstr(cstring);
	mu_assert(b != NULL, "Invalid b string");

	return NULL;
}

char *test_bdestroy()
{
	mu_assert(bdestroy(b) == BSTR_OK, "Bstring was not destroyed.");
	return NULL;
}

char *test_blk2bstr()
{
	b = blk2bstr(cstring, cstring_len);
	mu_assert(b != NULL, "Invalid b string");
	bdestroy(b);
	return NULL;
}

char *test_bstrcpy()
{
	b = bfromcstr(cstring);
	bstring b1 = bstrcpy(b);
	mu_assert(b1 != NULL, "Failed to copy bstring");
	mu_assert(bstrcmp(b, b1) == 0, "b1 is not a copy of b");
	bdestroy(b);
	bdestroy(b1);

	return NULL;
}

char *test_bassign()
{
	bstring a = bfromcstr("foo");
	b = bfromcstr(cstring);
	mu_assert(bassign(a, b) == BSTR_OK, "bassign method failed");
	mu_assert(bstrcmp(a, b) == 0, "String a was not overwritten");
	bdestroy(a);
	bdestroy(b);

	return NULL;
}

char *test_bassigncstr()
{
	bstring a = bfromcstr("foo");
	mu_assert(bassigncstr(a, "foobar") == BSTR_OK, "bassigncstr method failed");
	bdestroy(a);
	return NULL;
}

char *test_bassignblk()
{
	bstring a = bfromcstr("foo");
	b = bfromcstr(cstring);

	mu_assert(bassignblk(a, b->data, b->slen) == BSTR_OK, "bassignblk method failed");
	mu_assert(bstrcmp(a, b) == 0, "String a was not overwritten");
	bdestroy(a);
	bdestroy(b);

	return NULL;
}

char *test_bconcat()
{
	bstring a = bfromcstr("-foo");
	b = bfromcstr(cstring);
	char * concated = "foo-bar-baz-foo";
	bstring c = bfromcstr(concated);

	mu_assert(bconcat(b, a) == BSTR_OK, "bconcat method failed");
	mu_assert(bstrcmp(b, c) == 0, "Bstring was not concatenated");
	bdestroy(a);
	bdestroy(b);
	bdestroy(c);

	return NULL;
}

char *test_bstricmp()
{
	bstring a = bfromcstr("fOObAr");
	b = bfromcstr("FooBaR");

	mu_assert(a != NULL && b != NULL, "Bstring(s) were not successfully allocated.");
	mu_assert(bstricmp(a, b) == 0, "bstricmp method failed -- strings should be equal if ignoring case");
	bdestroy(a);
	bdestroy(b);

	return NULL;
}

char *test_biseq()
{
	bstring a = bfromcstr("foo");
	b = bfromcstr(cstring);
	bstring c = bfromcstr(cstring);

	mu_assert(biseq(a, b) == 0, "bstring a is different from bstring b");
	mu_assert(biseq(b, c) == 1, "bstring c is equal to bstring b");

	bdestroy_all(3, a, b, c);

	return NULL;
}

char *test_binstr()
{
	bstring a = bfromcstr("bar");
	b = bfromcstr(cstring);

	int rc = binstr(b, 0, a);

	mu_assert(rc != BSTR_ERR, "bstring a should be found in bstring b");
	mu_assert(rc == 4, "bstring a should be found at pos 4 in bstring b");
	mu_assert(binstr(b, 0, b) == 0, "same string is found at the 0th position");


	bdestroy_all(2, a, b);
	return NULL;
}

char *test_bfindreplace()
{
	bstring a = bfromcstr("a bar car tzar");
	bstring find = bfromcstr("a");
	bstring repl = bfromcstr("foo");
	bstring expected = bfromcstr("foo bfoor cfoor tzfoor");

	int rc = bfindreplace(a, find, repl, 0);

	mu_assert(rc == BSTR_OK, "bfindreplace failed");
	mu_assert(biseq(expected, a) == 1, "Replace didn't work correctly");

	bdestroy_all(4, a, find, repl, expected);

	return NULL;
}

char *test_bsplit()
{
	bstring to_split = bfromcstr("a bar car tzar");
	bstring a = bfromcstr("a");
	b = bfromcstr("bar");
	bstring c = bfromcstr("car");
	bstring d = bfromcstr("tzar");
	struct bstrList *list = bsplit(to_split, ' ');

	mu_assert(biseq(list->entry[0], a) == 1, "First list entry is incorrect");
	mu_assert(biseq(list->entry[1], b) == 1, "Second list entry is incorrect");
	mu_assert(biseq(list->entry[2], c) == 1, "Third list entry is incorrect");
	mu_assert(biseq(list->entry[3], d) == 1, "Last list entry is incorrect");

	bdestroy_all(5, to_split, a, b, c, d);
	bstrListDestroy(list);

	return NULL;
}

char *test_bformat()
{
	char *format = "%s is a #%u %s";
	char *a1 = "foo";
	char *a3 = "bar";
	unsigned int a2 = 3;

	char foo[15];
	sprintf(foo, format, a1, a2, a3);

	bstring res = bformat(format, a1, a2, a3);
	bstring expected = bfromcstr(foo);

	debug("Expected: %s", foo);
	debug("Result: %s", bdata(res));
	mu_assert(biseq(res, expected) == 1, "bstring produced an invalid result");


	bdestroy_all(2, res, expected);

	return NULL;
}

char *test_blength()
{
	bstring a = bfromcstr("Foo is a bar, baz?");

	mu_assert(blength(a) == 18, "bstring is different length than a string");
	mu_assert(blength(a) == a->slen, "Something really weird is going on.");

	bdestroy(a);

	return NULL;
}

char *test_bdata()
{
	bstring a = bfromcstr(cstring);

	mu_assert(strcmp(bdata(a), cstring) == 0, "Data is just imported cstring");

	bdestroy(a);

	return NULL;
}

char *test_bchar()
{
	bstring a = bfromcstr(cstring);

	mu_assert(bchar(a, 0) == 'f', "bchar grabbed the wrong letter");
	mu_assert(bchar(a, 100) == '\0', "a null character should be returned for an out-of-bounds index");

	bdestroy(a);

	return NULL;
}

char *all_tests()
{
	mu_suite_start();

	mu_run_test(test_bfromcstr);
	mu_run_test(test_bdestroy);
	mu_run_test(test_blk2bstr);
	mu_run_test(test_bstrcpy);
	mu_run_test(test_bassign);
	mu_run_test(test_bassigncstr);
	mu_run_test(test_bassignblk);
	mu_run_test(test_bconcat);
	mu_run_test(test_bstricmp);
	mu_run_test(test_biseq);
	mu_run_test(test_binstr);
	mu_run_test(test_bfindreplace);
	mu_run_test(test_bsplit);
	//mu_run_test(test_bformat);
	mu_run_test(test_blength);
	mu_run_test(test_bdata);
	mu_run_test(test_bchar);

	return NULL;
}

RUN_TESTS(all_tests)
